@@ -149,13 +149,11 @@ where
         initial_nodes: &[ConnectionInfo],
         retries: Option<u32>,
     ) -> RedisResult<Connection<C>> {
-        Pipeline::new(initial_nodes, retries)
-            .map_ok(|pipeline| {
-                let (tx, rx) = mpsc::channel::<Message<_>>(100);
-                tokio::spawn(rx.map(Ok).forward(pipeline).map(|_| ()));
-                Connection(tx)
-            })
-            .await
+        Pipeline::new(initial_nodes, retries).await.map(|pipeline| {
+            let (tx, rx) = mpsc::channel::<Message<_>>(100);
+            tokio::spawn(rx.map(Ok).forward(pipeline).map(|_| ()));
+            Connection(tx)
+        })
     }
 }
 
@@ -165,8 +163,11 @@ struct Pipeline<C> {
     connections: HashMap<String, C>,
     slots: SlotMap,
     state: ConnectionState<C>,
-    in_flight_requests:
-        Vec<Pin<Box<Request<BoxFuture<'static, (String, RedisResult<Response>)>, Response, C>>>>,
+    in_flight_requests: stream::FuturesUnordered<
+        Pin<Box<Request<BoxFuture<'static, (String, RedisResult<Response>)>, Response, C>>>,
+    >,
+    refresh_error: Option<RedisError>,
+    pending_requests: Vec<PendingRequest<Response, C>>,
     retries: Option<u32>,
 }
 
@@ -254,9 +255,12 @@ struct Message<C> {
     sender: oneshot::Sender<RedisResult<Response>>,
 }
 
+type RecoverFuture<C> =
+    BoxFuture<'static, Result<(SlotMap, HashMap<String, C>), (RedisError, HashMap<String, C>)>>;
+
 enum ConnectionState<C> {
     PollComplete,
-    Recover(RedisFuture<'static, (SlotMap, HashMap<String, C>)>),
+    Recover(RecoverFuture<C>),
 }
 
 impl<C> fmt::Debug for ConnectionState<C> {
@@ -293,40 +297,55 @@ pin_project! {
     }
 }
 
+struct PendingRequest<I, C> {
+    retry: u32,
+    sender: oneshot::Sender<RedisResult<I>>,
+    info: RequestInfo<C>,
+}
+
 pin_project! {
     struct Request<F, I, C> {
-        retry: u32,
         max_retries: Option<u32>,
-        sender: Option<oneshot::Sender<RedisResult<I>>>,
-        info: RequestInfo<C>,
+        request: Option<PendingRequest<I, C>>,
         #[pin]
         future: RequestState<F>,
     }
 }
 
 #[must_use]
-enum Next {
-    TryNewConnection,
+enum Next<I, C> {
+    TryNewConnection {
+        request: PendingRequest<I, C>,
+        error: Option<RedisError>,
+    },
+    Err {
+        request: PendingRequest<I, C>,
+        error: RedisError,
+    },
     Done,
 }
 
-impl<F, I, C> Request<F, I, C>
+impl<F, I, C> Future for Request<F, I, C>
 where
     F: Future<Output = (String, RedisResult<I>)>,
     C: ConnectionLike,
 {
-    fn poll_request(
-        mut self: Pin<&mut Self>,
-        cx: &mut task::Context,
-        connections_len: usize,
-    ) -> Poll<Result<Next, RedisError>> {
+    type Output = Next<I, C>;
+
+    fn poll(mut self: Pin<&mut Self>, cx: &mut task::Context) -> Poll<Self::Output> {
         let mut this = self.as_mut().project();
+        if this.request.is_none() {
+            return Poll::Ready(Next::Done);
+        }
         let future = match this.future.as_mut().project() {
             RequestStateProj::Future { future } => future,
             RequestStateProj::Sleep { sleep } => {
-                return Ok(match ready!(sleep.poll(cx)) {
-                    () => Next::TryNewConnection,
-                })
+                return match ready!(sleep.poll(cx)) {
+                    () => Next::TryNewConnection {
+                        request: self.project().request.take().unwrap(),
+                        error: None,
+                    },
+                }
                 .into();
             }
             _ => panic!("Request future must be Some"),
@@ -335,56 +354,68 @@ where
             (_, Ok(item)) => {
                 trace!("Ok");
                 self.respond(Ok(item));
-                Ok(Next::Done).into()
+                Next::Done.into()
             }
             (addr, Err(err)) => {
                 trace!("Request error {}", err);
 
-                match this.max_retries {
-                    Some(max_retries) if this.retry == max_retries => {
+                let request = this.request.as_mut().unwrap();
+
+                match *this.max_retries {
+                    Some(max_retries) if request.retry >= max_retries => {
                         self.respond(Err(err));
-                        return Ok(Next::Done).into();
+                        return Next::Done.into();
                     }
                     _ => (),
                 }
-                *this.retry = (*this.retry).saturating_add(1);
+                request.retry = request.retry.saturating_add(1);
 
                 if let Some(error_code) = err.code() {
                     if error_code == "MOVED" || error_code == "ASK" {
                         // Refresh slots and request again.
-                        this.info.excludes.clear();
-                        return Err(err).into();
+                        request.info.excludes.clear();
+                        return Next::Err {
+                            request: this.request.take().unwrap(),
+                            error: err,
+                        }
+                        .into();
                     } else if error_code == "TRYAGAIN" || error_code == "CLUSTERDOWN" {
                         // Sleep and retry.
                         let sleep_duration =
-                            Duration::from_millis(2u64.pow((*this.retry).max(7).min(16)) * 10);
-                        this.info.excludes.clear();
+                            Duration::from_millis(2u64.pow(request.retry.max(7).min(16)) * 10);
+                        request.info.excludes.clear();
                         this.future.set(RequestState::Sleep {
                             sleep: tokio::time::sleep(sleep_duration),
                         });
-                        return self.poll_request(cx, connections_len);
+                        return self.poll(cx);
                     }
                 }
 
-                this.info.excludes.insert(addr);
+                request.info.excludes.insert(addr);
 
-                if this.info.excludes.len() >= connections_len {
-                    self.respond(Err(err));
-                    return Ok(Next::Done).into();
+                Next::TryNewConnection {
+                    request: this.request.take().unwrap(),
+                    error: Some(err),
                 }
-
-                Ok(Next::TryNewConnection).into()
+                .into()
             }
         }
     }
+}
 
+impl<F, I, C> Request<F, I, C>
+where
+    F: Future<Output = (String, RedisResult<I>)>,
+    C: ConnectionLike,
+{
     fn respond(self: Pin<&mut Self>, msg: RedisResult<I>) {
         // If `send` errors the receiver has dropped and thus does not care about the message
         let _ = self
             .project()
-            .sender
+            .request
             .take()
             .expect("Result should only be sent once")
+            .sender
             .send(msg);
     }
 }
@@ -398,11 +429,13 @@ where
         let mut connection = Pipeline {
             connections,
             slots: Default::default(),
-            in_flight_requests: Vec::new(),
+            in_flight_requests: Default::default(),
+            refresh_error: None,
+            pending_requests: Vec::new(),
             state: ConnectionState::PollComplete,
             retries,
         };
-        let (slots, connections) = connection.refresh_slots().await?;
+        let (slots, connections) = connection.refresh_slots().await.map_err(|(err, _)| err)?;
         connection.slots = slots;
         connection.connections = connections;
         Ok(connection)
@@ -411,8 +444,8 @@ where
     async fn create_initial_connections(
         initial_nodes: &[ConnectionInfo],
     ) -> RedisResult<HashMap<String, C>> {
-        stream::iter(initial_nodes)
-            .then(|info| {
+        let connections = stream::iter(initial_nodes)
+            .map(|info| async move {
                 let addr = match *info.addr {
                     ConnectionAddr::Tcp(ref host, port) => match &info.passwd {
                         Some(pw) => format!("redis://:{}@{}:{}", pw, host, port),
@@ -421,11 +454,13 @@ where
                     _ => panic!("No reach."),
                 };
 
-                connect_and_check(info.clone()).map(|result| match result {
+                let result = connect_and_check(info.clone()).await;
+                match result {
                     Ok(conn) => Some((addr, conn)),
                     Err(_) => None,
-                })
+                }
             })
+            .buffer_unordered(initial_nodes.len())
             .fold(
                 HashMap::with_capacity(initial_nodes.len()),
                 |mut connections: HashMap<String, C>, conn: Option<(String, C)>| async move {
@@ -433,22 +468,21 @@ where
                     connections
                 },
             )
-            .map(|connections| {
-                if connections.len() == 0 {
-                    return Err(RedisError::from((
-                        ErrorKind::IoError,
-                        "Failed to create initial connections",
-                    )));
-                }
-                Ok(connections)
-            })
-            .await
+            .await;
+        if connections.len() == 0 {
+            return Err(RedisError::from((
+                ErrorKind::IoError,
+                "Failed to create initial connections",
+            )));
+        }
+        Ok(connections)
     }
 
     // Query a node to discover slot-> master mappings.
     fn refresh_slots(
         &mut self,
-    ) -> impl Future<Output = RedisResult<(SlotMap, HashMap<String, C>)>> {
+    ) -> impl Future<Output = Result<(SlotMap, HashMap<String, C>), (RedisError, HashMap<String, C>)>>
+    {
         let mut connections = mem::replace(&mut self.connections, Default::default());
 
         async move {
@@ -465,7 +499,10 @@ where
                     Err(err) => result = Err(err),
                 }
             }
-            let slots = result?;
+            let slots = match result {
+                Ok(slots) => slots,
+                Err(err) => return Err((err, connections)),
+            };
 
             // Remove dead connections and connect to new nodes if necessary
             let new_connections = HashMap::with_capacity(connections.len());
@@ -564,15 +601,118 @@ where
     ) -> impl Future<Output = (String, RedisResult<Response>)> {
         // TODO remove clone by changing the ConnectionLike trait
         let cmd = info.cmd.clone();
-        (if info.excludes.len() > 0 || info.slot.is_none() {
+        let f = if info.excludes.len() > 0 || info.slot.is_none() {
             future::Either::Left(future::ready(get_random_connection(
                 &self.connections,
                 Some(&info.excludes),
             )))
         } else {
             future::Either::Right(self.get_connection(info.slot.unwrap()))
-        })
-        .then(move |(addr, conn)| cmd.exec(conn).map(|result| (addr, result)))
+        };
+        async move {
+            let (addr, conn) = f.await;
+            let result = cmd.exec(conn).await;
+            (addr, result)
+        }
+    }
+
+    fn poll_recover(
+        &mut self,
+        cx: &mut task::Context<'_>,
+        mut future: RecoverFuture<C>,
+    ) -> Poll<Result<(), RedisError>> {
+        match future.as_mut().poll(cx) {
+            Poll::Ready(Ok((slots, connections))) => {
+                trace!("Recovered with {} connections!", connections.len());
+                self.slots = slots;
+                self.connections = connections;
+                self.state = ConnectionState::PollComplete;
+                Poll::Ready(Ok(()))
+            }
+            Poll::Pending => {
+                self.state = ConnectionState::Recover(future);
+                trace!("Recover not ready");
+                Poll::Pending
+            }
+            Poll::Ready(Err((err, connections))) => {
+                self.connections = connections;
+                self.state = ConnectionState::Recover(Box::pin(self.refresh_slots()));
+                Poll::Ready(Err(err))
+            }
+        }
+    }
+
+    fn poll_complete(&mut self, cx: &mut task::Context<'_>) -> Poll<Result<(), RedisError>> {
+        let mut connection_error = None;
+
+        if !self.pending_requests.is_empty() {
+            let mut pending_requests = mem::take(&mut self.pending_requests);
+            for request in pending_requests.drain(..) {
+                let future = self.try_request(&request.info);
+                self.in_flight_requests.push(Box::pin(Request {
+                    max_retries: self.retries,
+                    request: Some(request),
+                    future: RequestState::Future {
+                        future: future.boxed(),
+                    },
+                }));
+            }
+            self.pending_requests = pending_requests;
+        }
+
+        loop {
+            let result = match Pin::new(&mut self.in_flight_requests).poll_next(cx) {
+                Poll::Ready(Some(result)) => result,
+                Poll::Ready(None) | Poll::Pending => break,
+            };
+            let self_ = &mut *self;
+            match result {
+                Next::Done => {}
+                Next::TryNewConnection { request, error } => {
+                    if let Some(error) = error {
+                        if request.info.excludes.len() >= self_.connections.len() {
+                            let _ = request.sender.send(Err(error));
+                            continue;
+                        }
+                    }
+                    let future = self.try_request(&request.info);
+                    self.in_flight_requests.push(Box::pin(Request {
+                        max_retries: self.retries,
+                        request: Some(request),
+                        future: RequestState::Future {
+                            future: Box::pin(future),
+                        },
+                    }));
+                }
+                Next::Err { request, error } => {
+                    connection_error = Some(error);
+                    self.pending_requests.push(request);
+                }
+            }
+        }
+
+        if let Some(err) = connection_error {
+            Poll::Ready(Err(err))
+        } else if self.in_flight_requests.is_empty() {
+            Poll::Ready(Ok(()))
+        } else {
+            Poll::Pending
+        }
+    }
+
+    fn send_refresh_error(&mut self) {
+        if self.refresh_error.is_some() {
+            if let Some(mut request) = Pin::new(&mut self.in_flight_requests)
+                .iter_pin_mut()
+                .find(|request| request.request.is_some())
+            {
+                (*request)
+                    .as_mut()
+                    .respond(Err(self.refresh_error.take().unwrap()));
+            } else if let Some(request) = self.pending_requests.pop() {
+                let _ = request.sender.send(Err(self.refresh_error.take().unwrap()));
+            }
+        }
     }
 }
 
@@ -582,13 +722,37 @@ where
 {
     type Error = ();
 
-    fn poll_ready(self: Pin<&mut Self>, _cx: &mut task::Context) -> Poll<Result<(), Self::Error>> {
-        Ok(()).into()
+    fn poll_ready(
+        mut self: Pin<&mut Self>,
+        cx: &mut task::Context,
+    ) -> Poll<Result<(), Self::Error>> {
+        match mem::replace(&mut self.state, ConnectionState::PollComplete) {
+            ConnectionState::PollComplete => Poll::Ready(Ok(())),
+            ConnectionState::Recover(future) => {
+                match ready!(self.as_mut().poll_recover(cx, future)) {
+                    Ok(()) => Poll::Ready(Ok(())),
+                    Err(err) => {
+                        // We failed to reconnect, while we will try again we will report the
+                        // error if we can to avoid getting trapped in an infinite loop of
+                        // trying to reconnect
+                        if let Some(mut request) = Pin::new(&mut self.in_flight_requests)
+                            .iter_pin_mut()
+                            .find(|request| request.request.is_some())
+                        {
+                            (*request).as_mut().respond(Err(err));
+                        } else {
+                            self.refresh_error = Some(err);
+                        }
+                        Poll::Ready(Ok(()))
+                    }
+                }
+            }
+        }
     }
 
     fn start_send(mut self: Pin<&mut Self>, msg: Message<C>) -> Result<(), Self::Error> {
         trace!("start_send");
-        let cmd = msg.cmd;
+        let Message { cmd, sender } = msg;
 
         let excludes = HashSet::new();
         let slot = cmd.slot();
@@ -598,14 +762,12 @@ where
             slot,
             excludes,
         };
-        let request = Request {
-            max_retries: self.retries,
+
+        self.pending_requests.push(PendingRequest {
             retry: 0,
-            sender: Some(msg.sender),
-            future: RequestState::None,
+            sender,
             info,
-        };
-        self.in_flight_requests.push(Box::pin(request));
+        });
         Ok(()).into()
     }
 
@@ -615,87 +777,54 @@ where
     ) -> Poll<Result<(), Self::Error>> {
         trace!("poll_complete: {:?}", self.state);
         loop {
-            self.state = match mem::replace(&mut self.state, ConnectionState::PollComplete) {
-                ConnectionState::Recover(mut future) => match future.as_mut().poll(cx) {
-                    Poll::Ready(Ok((slots, connections))) => {
-                        trace!("Recovered with {} connections!", connections.len());
-                        self.slots = slots;
-                        self.connections = connections;
-                        ConnectionState::PollComplete
-                    }
-                    Poll::Pending => {
-                        self.state = ConnectionState::Recover(future);
-                        trace!("Recover not ready");
-                        return Poll::Pending;
-                    }
-                    Poll::Ready(Err(_err)) => {
-                        ConnectionState::Recover(Box::pin(self.refresh_slots()))
-                    }
-                },
-                ConnectionState::PollComplete => {
-                    let mut error = None;
-                    let mut i = 0;
+            self.send_refresh_error();
 
-                    while i < self.in_flight_requests.len() {
-                        if let RequestState::None = self.in_flight_requests[i].future {
-                            let future = self.try_request(&self.in_flight_requests[i].info);
-                            self.in_flight_requests[i].as_mut().project().future.set(
-                                RequestState::Future {
-                                    future: Box::pin(future),
-                                },
-                            );
+            match mem::replace(&mut self.state, ConnectionState::PollComplete) {
+                ConnectionState::Recover(future) => {
+                    match ready!(self.as_mut().poll_recover(cx, future)) {
+                        Ok(()) => (),
+                        Err(err) => {
+                            // We failed to reconnect, while we will try again we will report the
+                            // error if we can to avoid getting trapped in an infinite loop of
+                            // trying to reconnect
+                            self.refresh_error = Some(err);
+
+                            // Give other tasks a chance to progress before we try to recover
+                            // again. Since the future may not have registered a wake up we do so
+                            // now so the task is not forgotten
+                            cx.waker().wake_by_ref();
+                            return Poll::Pending;
                         }
-
-                        let self_ = &mut *self;
-                        match self_.in_flight_requests[i]
-                            .as_mut()
-                            .poll_request(cx, self_.connections.len())
-                        {
-                            Poll::Pending => {
-                                i += 1;
-                            }
-                            Poll::Ready(result) => match result {
-                                Ok(next) => match next {
-                                    Next::Done => {
-                                        self.in_flight_requests.swap_remove(i);
-                                    }
-                                    Next::TryNewConnection => {
-                                        let mut request = self.in_flight_requests.swap_remove(i);
-                                        let mut r = request.as_mut().project();
-                                        r.future.set(RequestState::Future {
-                                            future: Box::pin(self.try_request(&r.info)),
-                                        });
-                                        self.in_flight_requests.push(request);
-                                    }
-                                },
-                                Err(err) => {
-                                    error = Some(err);
-
-                                    self.in_flight_requests[i]
-                                        .as_mut()
-                                        .project()
-                                        .future
-                                        .set(RequestState::None);
-                                    i += 1;
-                                }
-                            },
-                        }
-                    }
-
-                    if let Some(err) = error {
-                        trace!("Recovering {}", err);
-                        ConnectionState::Recover(Box::pin(self.refresh_slots()))
-                    } else if self.in_flight_requests.is_empty() {
-                        return Ok(()).into();
-                    } else {
-                        return Poll::Pending;
                     }
                 }
+                ConnectionState::PollComplete => match ready!(self.poll_complete(cx)) {
+                    Ok(()) => return Poll::Ready(Ok(())),
+                    Err(err) => {
+                        trace!("Recovering {}", err);
+                        self.state = ConnectionState::Recover(Box::pin(self.refresh_slots()));
+                    }
+                },
             }
         }
     }
 
-    fn poll_close(self: Pin<&mut Self>, cx: &mut task::Context) -> Poll<Result<(), Self::Error>> {
+    fn poll_close(
+        mut self: Pin<&mut Self>,
+        cx: &mut task::Context,
+    ) -> Poll<Result<(), Self::Error>> {
+        // Try to drive any in flight requests to completion
+        match self.poll_complete(cx) {
+            Poll::Ready(result) => {
+                result.map_err(|_| ())?;
+            }
+            Poll::Pending => (),
+        };
+        // If we no longer have any requests in flight we are done (skips any reconnection
+        // attempts)
+        if self.in_flight_requests.is_empty() {
+            return Poll::Ready(Ok(()));
+        }
+
         self.poll_flush(cx)
     }
 }
@@ -714,19 +843,19 @@ where
                         cmd: Arc::new(cmd.clone()), // TODO Remove this clone?
                         func: |mut conn, cmd| {
                             Box::pin(async move {
-                                conn.req_packed_command(&cmd).map_ok(Response::Single).await
+                                conn.req_packed_command(&cmd).await.map(Response::Single)
                             })
                         },
                     },
                     sender,
                 })
+                .await
                 .map_err(|_| {
                     RedisError::from(io::Error::new(
                         io::ErrorKind::BrokenPipe,
                         "redis_cluster: Unable to send command",
                     ))
-                })
-                .await?;
+                })?;
             receiver
                 .await
                 .unwrap_or_else(|_| {
@@ -759,15 +888,15 @@ where
                         func: |mut conn, pipeline, offset, count| {
                             Box::pin(async move {
                                 conn.req_packed_commands(&pipeline, offset, count)
-                                    .map_ok(Response::Multiple)
                                     .await
+                                    .map(Response::Multiple)
                             })
                         },
                     },
                     sender,
                 })
-                .map_err(|_| RedisError::from(io::Error::from(io::ErrorKind::BrokenPipe)))
-                .await?;
+                .await
+                .map_err(|_| RedisError::from(io::Error::from(io::ErrorKind::BrokenPipe)))?;
 
             receiver
                 .await
@@ -916,13 +1045,10 @@ where
     trace!("get_slots");
     let mut cmd = Cmd::new();
     cmd.arg("CLUSTER").arg("SLOTS");
-    let value = connection
-        .req_packed_command(&cmd)
-        .map_err(|err| {
-            trace!("get_slots error: {}", err);
-            err
-        })
-        .await?;
+    let value = connection.req_packed_command(&cmd).await.map_err(|err| {
+        trace!("get_slots error: {}", err);
+        err
+    })?;
     trace!("get_slots -> {:#?}", value);
     // Parse response.
     let mut result = Vec::with_capacity(2);
